! module for storing data from ISW 
! real_Cl_file(with index,idata,il(lbinnumber),medianl,Cl,Cl_err) & 
! inv_covariance_matrix(in same indexing) Nparam 
! bdndz_file(to be output by bdndz code); 
! NOTE: order of the sample: 2MASS0 2MASS1 2MASS2 2MASS3 LRG0 LRG1 QSO0 QSO1 NVSS'

module iswdata  
  use ParamDef
  use CAMB
  use Settings
  use CMB_Cls
  use lrg_2dCl
  use cmbtypes
  use likelihood

  implicit none
  integer, parameter :: lmaxISW=600
  integer, parameter :: maxlist_par=1000, nmaxpar=15,max_len_matrix=100
  character(LEN=Ini_max_string_len) baserootISW, cwd,tmpdir !JD
  character(len=256) zfn
  character(len=256) cosmoparamfile
  character(len=1024) templine
  character(len=2048) command_system
  character(len=8) flag
  character(len=256) rootdndz
  
  
  type, extends(CosmologyLikelihood) :: ISWLikelihood
    contains
    procedure :: LogLike => ISWLnLike
  end type ISWLikelihood
  
  Type ISWdatapoint

    real,dimension(0:(max_len_matrix-1),0:5):: ISW_realCl
    real,dimension(0:(max_len_matrix-1),0:(max_len_matrix-1)):: ISW_inv_cov
    integer,dimension(0:lmaxISW) :: lsample
    

  end Type ISWdatapoint 

   
  type(ZStruct) :: zd

  integer:: im,jm
  integer:: io2,j
  real:: tmp0,tmp1,tmp2,tmp3,tmp4,tmp5,tmp6,mark
  integer:: cosmotype,idata
  real, dimension(1:Nsample,lmaxISW):: cllin
  integer:: ltmp,sizelsample,lindex,il,length_matrix
  Type(ISWdatapoint):: iswset

!initialization

contains
  subroutine ReadISWDataset(isw_filename)  
 
    character(LEN=Ini_max_string_len) isw_filename(5)
    character(len=80) realcl_file,inv_cov_file
    integer:: i,j,sizelsample,io2,mark
  
        if (feedback>0) write(*,*) "Initializing ISW-galaxy cross correlation data"
        
        call getcwd(cwd) !JD
        call get_environment_variable("TMPDIR",tmpdir)
        if (trim(tmpdir)=='')then
          tmpdir=trim(cwd)//'/bdndz_code/IO'
        end if
        
        !prepare names of files and commands fro bdndz code
        rootdndz = trim(baserootISW) 
        cosmotype = 1
        cosmoparamfile = trim(tmpdir)//'/'//trim(rootdndz)//'.dat'
        zfn = trim(tmpdir)//'/bdndz.'//trim(rootdndz)//'.dat'

        flag = '0'
        templine = 'cd '//trim(cwd)//'/bdndz_code/; ./dndz_scale.x'
        command_system =trim(templine)//' '//trim(zfn)//' '//trim(flag)//&
                        ' '//trim(cosmoparamfile)//'; cd ..'
        
        realcl_file = "data/real_Cl_1_k_0.05 "
        inv_cov_file = "data/inv_cov_full_1_k_0.05"
        ! inside ReadISWDataset: Reading ISW datafiles
        
        do i = 0,max_len_matrix-1 
            do j = 0, 5
                iswset%ISW_realCl(i,j) = 0.0
            enddo
        enddo

        !---reading in the real Cl file 
        iswset%lsample(0) = 0
        open(9,file=realcl_file, status='old')
        io2 = 0
        i = 0 
        do while (io2.eq.0) 
            read (9,*,iostat=io2) tmp0,tmp1,tmp2,tmp3,tmp4,tmp5
            if (io2 .ne. 0) cycle
            iswset%ISW_realCl(i,0) = tmp0 
            iswset%ISW_realCl(i,1) = tmp1 +1
            iswset%ISW_realCl(i,2) = tmp2 
            iswset%ISW_realCl(i,3) = tmp3 
            iswset%ISW_realCl(i,4) = tmp4 
            iswset%ISW_realCl(i,5) = tmp5
            i=i+1

! we are going to store all the  l that we would want for any sample 
! in an array called lsample
            mark = -1.0  !assume this l is not registered in this array before
            if ((abs(iswset%lsample(0)- 0)).gt. 0.01) then 
                sizelsample = iswset%lsample(0)
                do j =1,sizelsample    !lsample(0) is the size of the array
                    if (abs(iswset%ISW_realCl(i-1,3) - iswset%lsample(j)) .lt. 0.0001) then 
                        mark = 1.0 
                    endif
                enddo
                if (mark .lt. 0.0 ) then 
                    iswset%lsample(0) = iswset%lsample(0)+1
                    j = iswset%lsample(0)
                    iswset%lsample(j) = int(iswset%ISW_realCl(i-1,3))
                endif 
            endif

            if ((abs(iswset%lsample(0)-0)) .le. 0.01) then 
                iswset%lsample(0) = iswset%lsample(0)+1
                iswset%lsample(1) = int(iswset%ISW_realCl(i-1,3))
            endif
        enddo
        length_matrix = i
        close(9)

        !---- reading in the covariance matrix
        open(9,file=inv_cov_file, status='old')
        io2 = 0
        do while (io2.eq.0) 
            read (9,*,iostat=io2) tmp0,tmp1,tmp2
            if (io2 .ne. 0) cycle
            iswset%ISW_inv_cov(tmp0,tmp1) =  tmp2
        enddo
        close(9)

        if (length_matrix .ne. i) write(*,*) "WARNING: length of matrix is not consistent"
!---------------

  end subroutine ReadISWDataset


  real(mcp)  function IswLnLike(like,CMB,Theory,DataParams)
    Class(ISWLikelihood) :: like
    Class(CMBParams) CMB
    Class(TheoryPredictions) Theory
    real(mcp) DataParams(:)
    integer :: cosmotype,im,jm,il,iot
    !JD timing Failsafe variables ----------------
    integer, Dimension(13) ::buff
    integer :: status, cnt, rtime, zmtime, sz
    integer time, stat
    !-------------------------------------
    real :: omegam, TCMB,wlchi2
    real(dl),dimension(0:(max_len_matrix-1)):: tmp
    real(mcp):: chisq
   
        rtime = time()
        open(9,file=trim(cosmoparamfile),status='replace')
        omegam = CMB%omb + CMB%omc+ CMB%omnu
        TCMB = 2.726 ! This is set and used only in the ISW likelihood code, and it is
        ! consistent with what cosmomc uses in other parts of the code


        !writing to cosmoparam file that is read by the bdndz code
        write (9,'(I9)') cosmotype 
        write (9,'(E20.10)') TCMB
        write (9,'(E20.10)') CMB%omb
        write (9,'(E20.10)') CMB%omnu
        write (9,'(E20.10)') omegam
        write (9,'(E20.10)') CMB%omv
        write (9,'(E20.10)') CMB%w
        !currently the isw likelihood code doesn't support w_a (running of dark energy
        !equation of state), however the bdndz code does.
        write (9,'(E20.10)') 0.0   !w_a  
        write (9,'(E20.10)') (CMB%H0/100)
        write (9,'(E20.10)') Theory%sigma_8
        write (9,'(E20.10)') CMB%InitPower(1)  ! Lin%nn, scalar spectral index
        write (9,'(E20.10)') CMB%InitPower(3)! running of scalar spectral index
        !JD for Testing GR
        write (9,'(E20.10)') CMB%TGR_Q1
        write (9,'(E20.10)') CMB%TGR_D1
        write (9,'(E20.10)') CMB%TGR_Q2
        write (9,'(E20.10)') CMB%TGR_D2
        write (9,'(E20.10)') CMB%TGR_Q3
        write (9,'(E20.10)') CMB%TGR_D3
        write (9,'(E20.10)') CMB%TGR_Q4
        write (9,'(E20.10)') CMB%TGR_D4
        write (9,'(E20.10)') TGR_zdiv
        write (9,'(E20.10)') CMB%TGR_kc
        if(true_bin)then 
            write (9,'(I9)') 1
        else
            write (9,'(I9)') 0
        end if
        close(9)

        status = stat(trim(zfn),buff)
        zmtime = buff(10)
        call system(trim(command_system))


        !JD Timing Failsafe make sure zdistribution has been written completely
        cnt = 0
        Do While (cnt<3)
            status = stat(trim(zfn),buff)
            if (status == 0 .and. zmtime<buff(10) .and. buff(8)==sz) then
                cnt = cnt + 1
            end if
            sz = buff(8)
            if(time()>(rtime+40))then
                IswLnLike = logZero
                return
            end if
        end do
        !End JD timing failsafe

!-------- Read in zdistribution
        call lrg_rdzdist(zd, trim(zfn))
!Calling lrg_compute2dCl_Limber
        call lrg_compute2dCl_Limber(CMB,Theory, zd, lmaxISW, cllin,TCMB,iswset%lsample)

        cllin(:,:) = cllin(:,:)*1E6   ! want everything in microK !! 


!!---Now you have the theory Cls (in cllin) 
!!---Time to get the chisq !

!cllin(i) contains the Cl at l=i

!---taking (Cl_data-Cl_the)^T ( Cov^-1) (Cl_data-Cl_the)
!---one step at a time
!---1) Cov^-1(Cl_data-Cl_the)
!---2) (Cl_data-Cl_the)^T on the result of last step

        do im=0,(length_matrix-1) 
            tmp(im) = 0.0
        enddo 
        do im=0,(length_matrix-1) 
            do jm=0,(length_matrix-1) 
                ltmp = int(iswset%ISW_realCl(jm,3))
                idata = iswset%ISW_realCl(jm,1)

                sizelsample = iswset%lsample(0)
                do il = 1, sizelsample
                    if (ltmp .eq. iswset%lsample(il)) then
                        lindex = il
                    endif
                enddo
                tmp(im) = tmp(im) + iswset%ISW_inv_cov(im,jm)*(iswset%ISW_realCl(jm,4)&
                        -cllin(idata,lindex))
            enddo
        enddo
        chisq = 0.0

        do im=0,(length_matrix-1)  
            ltmp = int(iswset%ISW_realCl(im,3))
            idata =iswset%ISW_realCl(im,1)
            sizelsample = iswset%lsample(0)
            do il = 1, sizelsample
                if (ltmp .eq. iswset%lsample(il)) then
                    lindex = il
                endif
            enddo
            chisq = chisq+((iswset%ISW_realCl(im,4)-cllin(idata,lindex))*tmp(im))
        enddo

        IswLnLike = chisq/2.0    ! chisq/2 -> -ln(likelihood)

        if (Feedback > 1) write(*,*)'ISWlnlike = ', IswLnLike

  end function IswLnLike
  
  subroutine ISWLikelihood_Add(LikeList, Ini)
    use IniFile
    use settings
    class(LikelihoodList) :: LikeList
    Type(TIniFile) :: ini
    Type(ISWLikelihood), pointer :: like
    character (LEN=Ini_max_string_len):: isw_filename(5)
    
    use_ISW = Ini_Read_Logical_File(Ini, 'use_ISW',.false.)
    if (.not. use_ISW) return
    
    allocate(like)
    
    Like%LikelihoodType = 'ISW'
    Like%name='HoISW'
    like%needs_powerspectra = .true.
    like%max_z = 5
    like%num_z = 20
    call LikeList%Add(like)
    
    call ReadISWDataset(isw_filename)
  end subroutine ISWLikelihood_Add

end module iswdata
